import unittest
import unittest.mock
from unittest.mock import patch
from acrome import controller

class TestStewart(unittest.TestCase):
    def setUp(self) -> None:
        patcher = patch("acrome.controller.serial.Serial", autospec=True)
        self.mock = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock.reset_mock()
        self.dev = controller.Stewart()

    def tearDown(self):
        pass

    def test_enable(self):
        self.dev.enable(True)
        self.assertTrue(self.dev._Stewart__en)
        self.assertIsInstance(self.dev._Stewart__en, int)

        self.dev.enable(False)
        self.assertFalse(self.dev._Stewart__en)
        self.assertIsInstance(self.dev._Stewart__en, int)
        
        self.dev.enable(3)
        self.assertTrue(self.dev._Stewart__en)
        self.assertIsInstance(self.dev._Stewart__en, int)

        self.dev.enable(4)
        self.assertFalse(self.dev._Stewart__en)
        self.assertIsInstance(self.dev._Stewart__en, int)

    def test_set_motors_valid_values(self):
        for mt in range(-self.dev.__class__._MAX_MT_ABS, self.dev.__class__._MAX_MT_ABS+1):
            self.dev.set_motors([mt] * 6)
            self.assertEqual(self.dev._Stewart__motors, [mt] * 6)
        
        for m in self.dev._Stewart__motors:
            self.assertIsInstance(m, int)
        
    def test_set_motors_invalid_values(self):
        self.dev.set_motors([99999] * 6)
        self.assertEqual(self.dev._Stewart__motors, [self.dev.__class__._MAX_MT_ABS] * 6)
        for m in self.dev._Stewart__motors:
            self.assertIsInstance(m, int)

        self.dev.set_motors([-99999] * 6)
        self.assertEqual(self.dev._Stewart__motors, [-self.dev.__class__._MAX_MT_ABS] * 6)
        
        for m in self.dev._Stewart__motors:
            self.assertIsInstance(m, int)

    def test_write(self):
        self.dev.enable(True)
        self.dev.set_motors([100,-200,300,-400,500,-600])
        
        with patch.object(controller.Controller, '_writebus') as wr:
            self.dev._write()
        
        wr.assert_called_once_with(bytes([0x55, 0xBE, 0x1, 0x64, 0x0, 0x38, 0xFF, 0x2C, 0x1, 0x70, 0xFE, 0xF4, 0x1, 0xA8, 0xFD, 0x52, 0x9A, 0xEF, 0xB7]))

    def test_read(self):
        #POS 50,63,85,117,756,3721
        self.mock.return_value.read.return_value = bytes([0x55, 0xBE, 0x32, 0x0, 0x3F, 0x0, 0x55, 0x0, 0x75, 0x0, 0xF4, 0x2, 0x89, 0xE, 0x9A, 0x99, 0xCC, 0x42, 0x9A, 0x19, 0x59, 0x43, 0x33, 0xF3, 0xB3, 0x43, 0x90, 0xC5, 0xF0, 0x59])
    
        self.dev._read()

        self.assertEqual(self.dev.position, [50,63,85,117,756,3721])
        for i, val in enumerate([102.3, 217.1, 359.9]):
            self.assertAlmostEqual(self.dev.imu[i], val, places=2)

    def test_update(self):
        
        with patch.object(self.dev.__class__, '_write') as wr:
            self.dev.update()
            wr.assert_called()

        with patch.object(self.dev.__class__, '_read') as rd:
            self.dev.update()
            rd.assert_called()

    def test_reboot(self):
        with patch.object(controller.Controller, '_writebus') as wr:
            self.dev.reboot()
            wr.assert_called_once_with(bytes([0x55, 0xFC, 0x1, 0x0, 0x0, 0x0, 0x0, 0xA3, 0x41, 0x95, 0xD2]))
    
    def test_enter_bootloader(self):
        with patch.object(controller.Controller, '_writebus') as wr:
            self.dev.enter_bootloader()
            wr.assert_called_once_with(bytes([0x55, 0xFC, 0x2, 0x0, 0x0, 0x0, 0x0, 0x34, 0xE9, 0x82, 0x9]))

    def test_ping(self):
        self.mock.return_value.read.return_value = bytes([0x55, 0x0, 0x57, 0x73, 0x9D, 0xC6])
        with patch.object(controller.Controller, '_writebus') as wr:
            self.assertTrue(self.dev.ping())
            wr.assert_called_once_with(bytes([0x55, 0x0, 0x57, 0x73, 0x9D, 0xC6]))

    def test_get_status(self):
        self.mock.return_value.read.return_value = bytes([0x55, 0xFC, 0x0, 0x1, 0x0, 0x0, 0x0, 0x1, 0x1, 0x0, 0x15, 0x0, 0x0, 0x0, 0x0, 0xF1, 0x79, 0xD6, 0x6F])
        with patch.object(controller.Controller, '_writebus') as wr:
            st = self.dev.get_board_info()
            wr.assert_called_once_with(bytes([0x55, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2E, 0x26, 0x98, 0x9B]))
            
        


import serial.tools.list_ports
import platform

def whichOS():
    return platform.system()


def USB_serial_port():

    if whichOS() == "Windows":
        ports = list(serial.tools.list_ports.comports())
        if ports:
            for port, desc, hwid in sorted(ports):
                #print(f"{port}: {desc} [{hwid}]")
                #print(type(port))
                if 'USB Serial Port' in desc:
                    return port
        else:
            return None
        
    if whichOS() == "Linux":
        ports = list(serial.tools.list_ports.comports())
        if ports:
            for port, desc, hwid in sorted(ports):
                #print(f"{port}: {desc} [{hwid}]")
                #print(type(port))
                if '/dev/ttyUSB' in port:
                    return port
        else:
            return None



import unittest
import unittest.mock
from unittest.mock import patch
from acrome import controller

class TestDelta(unittest.TestCase):
    def setUp(self) -> None:
        patcher = patch("acrome.controller.serial.Serial", autospec=True)
        self.mock = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock.reset_mock()
        self.dev = controller.Delta()

    def tearDown(self):
        pass

    def test_set_motors_valid_values(self):
        for mt in range(self.dev.__class__._MIN_MT_POS, self.dev.__class__._MAX_MT_POS):
            self.dev.set_motors([mt] * 3)
            self.assertEqual(self.dev._Delta__motors, [mt] * 3)
            for m in self.dev._Delta__motors:
                self.assertIsInstance(m, int)
        
    def test_set_motors_invalid_values(self):
        self.dev.set_motors([99999] * 3)
        self.assertEqual(self.dev._Delta__motors, [self.dev.__class__._MAX_MT_POS] * 3)
        for m in self.dev._Delta__motors:
            self.assertIsInstance(m, int)
        
        self.dev.set_motors([-99999] * 3)
        self.assertEqual(self.dev._Delta__motors, [self.dev.__class__._MIN_MT_POS] * 3)
        for m in self.dev._Delta__motors:
            self.assertIsInstance(m, int)

    def test_pick(self):
        self.dev.pick(True)
        self.assertTrue(self.dev._Delta__magnet == 1)
        self.dev.pick(False)
        self.assertTrue(self.dev._Delta__magnet == 0)

    def test_write(self):
        self.dev.pick(True)
        self.dev.set_motors([400, 500, 600])
        
        with patch.object(controller.Controller, '_writebus') as wr:
            self.dev._write()
        
        wr.assert_called_once_with(bytes([0x55, 0xBD, 0x1, 0x90, 0x1, 0xF4, 0x1, 0x58, 0x2, 0x27, 0xA7, 0x2C, 0x7A]))                     

    def test_read(self):
        #POS 317,656,1721
        self.mock.return_value.read.return_value = bytes([0x55, 0xBD, 0x3D, 0x1, 0x90, 0x2, 0xB9, 0x6, 0x1D, 0xCB, 0x83, 0xD6])
    
        self.dev._read()

        self.assertEqual(self.dev.position, [317,656,1721])
    
    def test_update(self):
        
        with patch.object(self.dev.__class__, '_write') as wr:
            self.dev.update()
            wr.assert_called()

        with patch.object(self.dev.__class__, '_read') as rd:
            self.dev.update()
            rd.assert_called()
    
    def test_reboot(self):
        with patch.object(controller.Controller, '_writebus') as wr:
            self.dev.reboot()
            wr.assert_called_once_with(bytes([0x55, 0xFC, 0x1, 0x0, 0x0, 0x0, 0x0, 0xA3, 0x41, 0x95, 0xD2]))
    
    def test_enter_bootloader(self):
        with patch.object(controller.Controller, '_writebus') as wr:
            self.dev.enter_bootloader()
            wr.assert_called_once_with(bytes([0x55, 0xFC, 0x2, 0x0, 0x0, 0x0, 0x0, 0x34, 0xE9, 0x82, 0x9]))

    def test_ping(self):
        self.mock.return_value.read.return_value = bytes([0x55, 0x0, 0x57, 0x73, 0x9D, 0xC6])
        with patch.object(controller.Controller, '_writebus') as wr:
            self.assertTrue(self.dev.ping())
            wr.assert_called_once_with(bytes([0x55, 0x0, 0x57, 0x73, 0x9D, 0xC6]))

    def test_get_status(self):
        self.mock.return_value.read.return_value = bytes([0x55, 0xFC, 0x0, 0x1, 0x0, 0x0, 0x0, 0x1, 0x1, 0x0, 0x15, 0x0, 0x0, 0x0, 0x0, 0xF1, 0x79, 0xD6, 0x6F])
        with patch.object(controller.Controller, '_writebus') as wr:
            st = self.dev.get_board_info()
            wr.assert_called_once_with(bytes([0x55, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2E, 0x26, 0x98, 0x9B]))



import enum
import struct
from crccheck.crc import Crc32Mpeg2 as CRC32
import time
import serial

Index = enum.IntEnum('Index', [
	'Header',
	'DeviceID',
	'PackageSize',
	'Command',
	'SoftwareVersion',
	'HardwareVersion',
	# //////
	'ErrorCount',
	'CurrentPosition',
	'CurrentSpeed',
	'LimitSwitchStatus',
	'BusVoltage',
	'V5Voltage',
	# //// Position Control Parameters //////
	'MinPosition',
	'MaxPosition',
	'MinSpeed_POS',
	'MaxSpeed_POS',
	'MaxAccel_POS',
	'GoalPosition',
	'EncoderDeadband_POS',
	'Trajectory_time',
	'Trajectory_accel',
	'Trajectory_maxSpeed',
	'LimitSwitch_enable',
	# ////// Velocity Control Parameters //////
	'MaxSpeed_VELOCITY',
	'GoalSpeed',
	'Acceleration_VELOCITY',
	# ////////// General /////////////////////
	'StepperRunMode',
	'IsEncoderFeedBack',
	'Baudrate',
	'TorqueEn',
	'PIOMode0',
	'PIOMode1',
	'PIOMode2',
	'PIOMode3',
	'PIOData0',
	'PIOData1',
	'PIOData2',
	'PIOData3',
	'ExternalEncoder',
	'TestValue_i_1',
	'TestValue_i_2',
	'TestValue_i_3',
	'TestValue_i_4',
	'TestValue_f_1',
	'TestValue_f_2',
	'TestValue_f_3',
	'TestValue_f_4',
	'TestValue_b_1', 
	'TestValue_b_2',
	'TestValue_b_3', 
	'TestValue_b_4',
	'CRCValue',
], start=0)

class Commands(enum.IntEnum):
	PING = 0x00,
	READ = 0x01,
	WRITE = 0x02,
	REBOOT = 0x10,
	EEPROM_WRITE = 0x20,
	CALIBRATE = 0x25,
	BL_JUMP = 0x30,
	WRITE_SYNC = 0x40,
	READ_SYNC = 0x41,
	SLOW_STOP = 0x60,
	STOP = 0x61,
	ACK = 0x80,
	WRITE_ACK = 0x80 | 0x02,
	EEPROM_WRITE_ACK = 0x20 | 0x02,
	
class _Data():
	def __init__(self, index, var_type, rw=True, value = 0):
		self.__index = index
		self.__type = var_type
		self.__size  = struct.calcsize(self.__type)
		self.__value = value
		self.__rw = rw

	def value(self, value=None):
		if value is None:
			return self.__value
		elif self.__rw:
			self.__value = struct.unpack('<' + self.__type, struct.pack('<' + self.__type, value))[0]

	def index(self) ->enum.IntEnum:
		return self.__index

	def size(self) -> int:
		return self.__size
	
	def type(self) -> str:
		return self.__type

class Blue():
	_BATCH_ID = 254
	def __init__(self, ID) -> bool:
		self.__ack_size = 0
		self.vars = [
			_Data(Index.Header, 'B', False, 0x55),
			_Data(Index.DeviceID, 'B'),
			_Data(Index.PackageSize, 'B'),
			_Data(Index.Command, 'B'),
			_Data(Index.SoftwareVersion, 'I'),
			_Data(Index.HardwareVersion, 'I'),
			_Data(Index.ErrorCount, 'I'),
			_Data(Index.CurrentPosition, 'i'),
			_Data(Index.CurrentSpeed, 'I'),
			_Data(Index.LimitSwitchStatus, 'B'),
			_Data(Index.BusVoltage, 'H'),
			_Data(Index.V5Voltage, 'H'),
			# ////////////////////////////////////////
			_Data(Index.MinPosition,'i'),
			_Data(Index.MaxPosition,'i'),
			_Data(Index.MinSpeed_POS,'i'),
			_Data(Index.MaxSpeed_POS,'i'),
			_Data(Index.MaxAccel_POS,'i'),
			_Data(Index.GoalPosition,'i'),
			_Data(Index.EncoderDeadband_POS,'I'),
			_Data(Index.Trajectory_time,'f'),
			_Data(Index.Trajectory_accel,'f'),
			_Data(Index.Trajectory_maxSpeed,'f'),
			_Data(Index.LimitSwitch_enable,'B'),
			# ////// Velocity Control Parameters //////
			_Data(Index.MaxSpeed_VELOCITY, 'i'),
			_Data(Index.GoalSpeed, 'i'),
			_Data(Index.Acceleration_VELOCITY,'i'),
			# ////////// General /////////////////////
			_Data(Index.StepperRunMode, 'B'),
			_Data(Index.IsEncoderFeedBack, 'B'),
			_Data(Index.Baudrate, 'I'),
			_Data(Index.TorqueEn, 'B'),
			_Data(Index.PIOMode0, 'B'),
			_Data(Index.PIOMode1, 'B'),
			_Data(Index.PIOMode2, 'B'),
			_Data(Index.PIOMode3, 'B'),
			_Data(Index.PIOData0, 'I'),
			_Data(Index.PIOData1, 'I'),
			_Data(Index.PIOData2, 'I'),
			_Data(Index.PIOData3, 'I'),
			_Data(Index.ExternalEncoder, 'I'),
			_Data(Index.TestValue_i_1, 'i'),
			_Data(Index.TestValue_i_2, 'i'),
			_Data(Index.TestValue_i_3, 'i'),
			_Data(Index.TestValue_i_4, 'i'),
			_Data(Index.TestValue_f_1, 'f'),
			_Data(Index.TestValue_f_2, 'f'),
			_Data(Index.TestValue_f_3, 'f'),
			_Data(Index.TestValue_f_4, 'f'),
			_Data(Index.TestValue_b_1, 'b'),
			_Data(Index.TestValue_b_2, 'b'),
			_Data(Index.TestValue_b_3, 'b'),
			_Data(Index.TestValue_b_4, 'b'),
			_Data(Index.CRCValue, 'I')
		]

		if ID > 255 or ID < 0:
			raise ValueError("Device ID can not be higher than 253 or lower than 0!")
		else:
			self.vars[Index.DeviceID].value(ID)

	def get_ack_size(self):
		return self.__ack_size

	def set_variables(self, index_list=[], value_list=[], ack=False):
		#Set command to write/write_ack
		self.vars[Index.Command].value(Commands.WRITE_ACK if ack else Commands.WRITE)
		
		fmt_str = '<' + ''.join([var.type() for var in self.vars[:4]])
		for index, value in zip(index_list, value_list):
			self.vars[int(index)].value(value)
			fmt_str += 'B' + self.vars[int(index)].type()
		self.__ack_size = struct.calcsize(fmt_str)
		#Create a list of id-value pairs and convert them to a byte-array
		struct_out = list(struct.pack(fmt_str, *[*[var.value() for var in self.vars[:4]], *[val for pair in zip(index_list, [self.vars[int(index)].value() for index in index_list]) for val in pair]]))
		struct_out[int(Index.PackageSize)] = len(struct_out) + self.vars[Index.CRCValue].size()
		self.vars[Index.CRCValue].value(CRC32.calc(struct_out))		
		return bytes(struct_out) + struct.pack('<' + self.vars[Index.CRCValue].type(), self.vars[Index.CRCValue].value())
	
	def get_variables(self, index_list=[]):
		#Set command to Read
		self.vars[Index.Command].value(Commands.READ)
		
		#Constant Registers
		fmt_str = '<' + ''.join([var.type() for var in self.vars[:4]])
		
		#Update format string
		for _ in index_list:
			fmt_str += 'B'

		self.__ack_size = struct.calcsize(fmt_str + self.vars[Index.CRCValue].type()) + struct.calcsize(''.join([self.vars[idx].type() for idx in index_list]))

		#Populate actual string
		struct_out = list(struct.pack(fmt_str, *[*[var.value() for var in self.vars[:4]], *[int(index) for index in index_list]]))
		
		#Populate actual package size
		struct_out[int(Index.PackageSize)] = len(struct_out) + self.vars[Index.CRCValue].size()
		
		#Calculate CRC
		self.vars[Index.CRCValue].value(CRC32.calc(struct_out))

		#Append CRC and return
		return bytes(struct_out) + struct.pack('<' + self.vars[Index.CRCValue].type(), self.vars[Index.CRCValue].value())
	
	def reboot(self):
		self.vars[Index.Command].value(Commands.REBOOT)
		fmt_str = '<' + ''.join([var.type() for var in self.vars[:4]])
		struct_out = list(struct.pack(fmt_str, *[var.value() for var in self.vars[:4]]))
		struct_out[int(Index.PackageSize)] = len(struct_out) + self.vars[Index.CRCValue].size()
		self.vars[Index.CRCValue].value(CRC32.calc(struct_out))
		self.__ack_size = 0
		return bytes(struct_out) + struct.pack('<' + self.vars[Index.CRCValue].type(), self.vars[Index.CRCValue].value())
	
	def EEPROM_write(self, ack=False):
		self.vars[Index.Command].value(Commands.EEPROM_WRITE_ACK if ack else Commands.EEPROM_WRITE)
		fmt_str = '<' + ''.join([var.type() for var in self.vars[:4]])
		struct_out = list(struct.pack(fmt_str, *[var.value() for var in self.vars[:4]]))
		struct_out[int(Index.PackageSize)] = len(struct_out) + self.vars[Index.CRCValue].size()
		self.vars[Index.CRCValue].value(CRC32.calc(struct_out))
		self.__ack_size = struct.calcsize(fmt_str + self.vars[Index.CRCValue].type())
		return bytes(struct_out) + struct.pack('<' + self.vars[Index.CRCValue].type(), self.vars[Index.CRCValue].value())
	
	def ping(self):
		self.vars[Index.Command].value(Commands.PING)
		fmt_str = '<' + ''.join([var.type() for var in self.vars[:4]])
		struct_out = list(struct.pack(fmt_str, *[var.value() for var in self.vars[:4]]))
		struct_out[int(Index.PackageSize)] = len(struct_out) + self.vars[Index.CRCValue].size()
		self.vars[Index.CRCValue].value(CRC32.calc(struct_out))
		self.__ack_size = struct.calcsize(fmt_str + self.vars[Index.CRCValue].type())
		return bytes(struct_out) + struct.pack('<' + self.vars[Index.CRCValue].type(), self.vars[Index.CRCValue].value())

	def slow_stop(self):
		self.vars[Index.Command].value(Commands.SLOW_STOP)
		fmt_str = '<' + ''.join([var.type() for var in self.vars[:4]])
		struct_out = list(struct.pack(fmt_str, *[var.value() for var in self.vars[:4]]))
		struct_out[int(Index.PackageSize)] = len(struct_out) + self.vars[Index.CRCValue].size()
		self.vars[Index.CRCValue].value(CRC32.calc(struct_out))
		self.__ack_size = struct.calcsize(fmt_str + self.vars[Index.CRCValue].type())
		return bytes(struct_out) + struct.pack('<' + self.vars[Index.CRCValue].type(), self.vars[Index.CRCValue].value())
	
	def calibrate(self):
		self.vars[Index.Command].value(Commands.CALIBRATE)
		fmt_str = '<' + ''.join([var.type() for var in self.vars[:4]])
		struct_out = list(struct.pack(fmt_str, *[var.value() for var in self.vars[:4]]))
		struct_out[int(Index.PackageSize)] = len(struct_out) + self.vars[Index.CRCValue].size()
		self.vars[Index.CRCValue].value(CRC32.calc(struct_out))
		self.__ack_size = 0
		return bytes(struct_out) + struct.pack('<' + self.vars[Index.CRCValue].type(), self.vars[Index.CRCValue].value())


class Master():
	def __init__(self, portname, baudrate=115200):
		self.__driver_list = [Blue(255)] * 256
		if baudrate > 9500000 or baudrate < 1200:
			raise ValueError("Baudrate must be in range of 1200 to 9.5M")
		else:
			self.__baudrate = baudrate
			self.__post_sleep = 0
			self.__ph = serial.Serial(port=portname, baudrate=self.__baudrate, timeout=0.1)

	def __del__(self):
		try:
			self.__ph.close()
		except:
			pass

	def __write_bus(self, data):
		self.__ph.write(data)

	def __read_bus(self, size) -> bytes:
		self.__ph.flushInput()
		return self.__ph.read(size=size)

	def attach(self, blue:Blue):
		self.__driver_list[blue.vars[Index.DeviceID].value()] = blue
	
	def busy_wait(self, wait_time):
		start_time = time.time()
		while (time.time() - start_time) < wait_time:
			pass

	def parse_received(self, data):
		id = data[Index.DeviceID]
		data = data[4:-4]
		fmt_str = '<'

		i = 0
		while i < len(data):
			fmt_str += 'B' + self.__driver_list[id].vars[data[i]].type()
			i += self.__driver_list[id].vars[data[i]].size() + 1

		unpacked = list(struct.unpack(fmt_str, data))
		grouped = zip(*(iter(unpacked),) * 2)
		for group in grouped:
			self.__driver_list[id].vars[group[0]].value(group[1])

	def set_variables(self, id, idx_val_pairs=[], ack=False) -> list:
		index_list = [pair[0] for pair in idx_val_pairs]
		value_list = [pair[1] for pair in idx_val_pairs]

		self.__write_bus(self.__driver_list[id].set_variables(index_list, value_list, ack))
		if ack:
			if self.__read_ack(id):
				return [self.__driver_list[id].vars[index].value() for index in index_list]
		#time.sleep(self.__post_sleep)
		return [None]

	def get_variables(self, id, index_list) -> list:
		self.__write_bus(self.__driver_list[id].get_variables(index_list))
		if self.__read_ack(id):
			return [self.__driver_list[id].vars[index].value() for index in index_list]
		else:
			return [None]

	def __read_ack(self, id) -> bool:
		ret = self.__read_bus(self.__driver_list[id].get_ack_size())
		if len(ret) == self.__driver_list[id].get_ack_size():
			if (CRC32.calc(ret[:-4]) == struct.unpack('<I', ret[-4:])[0]):
				if ret[int(Index.PackageSize)] > 8:
					self.parse_received(ret)
					return True
				else:
					return True
			else:
				return False
		else:
			return False

	def reboot(self, id):
		self.__write_bus(self.__driver_list[id].reboot())
		time.sleep(self.__post_sleep)

	def EEPROM_write(self, id, ack=False):
		index_list = [int(index) for index in Index]
		self.__write_bus(self.__driver_list[id].EEPROM_write(ack))
		if ack:
			if self.__read_ack(id):
				return [self.__driver_list[id].vars[index].value() for index in index_list]
		time.sleep(self.__post_sleep)
		return [None]

	def ping(self, id):
		self.__write_bus(self.__driver_list[id].ping())
		if self.__read_ack(id):
			return True
		
	def calibrate(self, id):
		self.__write_bus(self.__driver_list[id].calibrate())
		time.sleep(self.__post_sleep)
		
	def slow_stop(self, id):
		self.__write_bus(self.__driver_list[id].slow_stop())
		time.sleep(self.__post_sleep)


	def set_variable_combined(self, val_indexes, values_lists, device_size ,ack=False) -> list:
		"""		
			- val_indexes should be list [val_index1, val_index2, ...]
			- values should be list in list [[value1, value2, ...], [value1, value2, ...], ...]
		--------> Ex Usage <-------------------------------------------------------------------------

		set_variable_combined([Index.GoalPosition, Index.Trajectory_time], [[1, 2, 3], [4, 5, 6]], 3)

			------ (this function sets 2 value of 3 devices.)

		---------------------------------------------------------------------------------------------
		"""

		if(len(val_indexes) != len(values_lists)):
			raise ValueError("val_indexes and values_lists should have same length")
		for value_list in values_lists:
			if(len(value_list) != device_size):
				raise ValueError(" The length of all of value_lists must be as long as the device size. ")
		#Set command
		self.__driver_list[255].vars[Index.Command].value(Commands.WRITE_SYNC)

		fmt_str = '<' + 'BBBBB'
		for val_index in val_indexes:
			fmt_str = fmt_str + 'B' + (self.__driver_list[255].vars[val_index]).type()*device_size

		indexes_and_values_together = []
		for val_index, value_list in zip(val_indexes, values_lists):
			indexes_and_values_together.append(val_index)
			for value in value_list:
				indexes_and_values_together.append(value)
		
		struct_out = list(struct.pack(fmt_str, *[var.value() for var in self.__driver_list[255].vars[:4]], device_size ,*indexes_and_values_together))

		struct_out[int(Index.PackageSize)] = len(struct_out) + self.__driver_list[255].vars[Index.CRCValue].size()						# editing package size
		self.__driver_list[255].vars[Index.CRCValue].value(CRC32.calc(struct_out))														# adding CRC

		readyToSend = bytes(struct_out) + struct.pack('<' + self.__driver_list[255].vars[Index.CRCValue].type(), self.__driver_list[255].vars[Index.CRCValue].value())
		self.__write_bus(readyToSend)
		return [None]




import unittest
import unittest.mock
from unittest.mock import patch
from acrome import controller

class TestOneDOF(unittest.TestCase):
    def setUp(self) -> None:
        patcher = patch("acrome.controller.serial.Serial", autospec=True)
        self.mock = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock.reset_mock()
        self.dev = controller.OneDOF()

    def tearDown(self):
        self.dev.__init__()

    def test_set_speed_valid_values(self):
        for speed in range(-1000,1000+1):
            self.dev.set_speed(speed)
            self.assertEqual(self.dev._OneDOF__speed, speed)

        self.assertTrue(isinstance(self.dev._OneDOF__speed, int))

    def test_set_speed_invalid_values(self):
        self.dev.set_speed(99999999)
        self.assertEqual(self.dev._OneDOF__speed, self.dev.__class__._MAX_SPEED_ABS)
        
        self.assertTrue(isinstance(self.dev._OneDOF__speed, int))

        self.dev.set_speed(-99999999)
        self.assertEqual(self.dev._OneDOF__speed, -self.dev.__class__._MAX_SPEED_ABS)

        self.assertTrue(isinstance(self.dev._OneDOF__speed, int))

    def test_set_enable(self):
        first_config = self.dev._OneDOF__config
        self.dev.enable(1)
        self.assertEqual(self.dev._OneDOF__config, first_config | self.dev.__class__._EN_MASK)
        self.assertTrue(isinstance(self.dev._OneDOF__config, int))

    def test_reset_enable(self):
        self.dev._OneDOF__config |= self.dev._EN_MASK
        first_config = self.dev._OneDOF__config
        self.dev.enable(False)
        self.assertEqual(self.dev._OneDOF__config, first_config & ~self.dev._EN_MASK)
        self.assertTrue(isinstance(self.dev._OneDOF__config, int))
        
    def test_reset_encoder(self):
        first_config = self.dev._OneDOF__config
        self.dev.reset_encoder_mt()
        self.assertEqual(self.dev._OneDOF__config, first_config | self.dev._ENC1_RST_MASK)
        self.assertTrue(isinstance(self.dev._OneDOF__config, int))
        self.dev._OneDOF__config = 0
        
        
        first_config = self.dev._OneDOF__config
        self.dev.reset_encoder_shaft()
        self.assertEqual(self.dev._OneDOF__config, first_config | self.dev._ENC2_RST_MASK)
        self.assertTrue(isinstance(self.dev._OneDOF__config, int))
    
    def test_write(self):
        self.dev.enable(1)
        self.dev.set_speed(500)
        self.dev.reset_encoder_mt()
        
        with patch.object(controller.Controller, '_writebus') as wr:
            self.dev._write()
        
        wr.assert_called_once_with(bytes([0x55, 0xBA, 0x03, 0xF4, 0x01, 0x2A, 0x64, 0xAE, 0xE5]))

        self.assertEqual(self.dev._OneDOF__config, self.dev._EN_MASK)
        
    def test_read(self):
        #ENC1 27381
        #ENC2 1998
        #IMU 117.4, 32.2, 258.8
        self.mock.return_value.read.return_value = bytes([0x55, 0xBA, 0xF5, 0x6A, 0xCE, 0x7, 0xCD, 0xCC, 0xEA, 0x42, 0xCD, 0xCC, 0x0, 0x42, 0x66, 0x66, 0x81, 0x43, 0xF1, 0x92, 0x3C, 0xCE])

        self.dev._read()

        self.assertEqual(self.dev.motor_enc, 27381)
        self.assertEqual(self.dev.shaft_enc, 1998)
        for i, val in enumerate([117.4, 32.2, 258.8]):
            self.assertAlmostEqual(self.dev.imu[i], val, places=2)
    
    def test_update(self):
        
        with patch.object(self.dev.__class__, '_write') as wr:
            self.dev.update()
            wr.assert_called()

        with patch.object(self.dev.__class__, '_read') as rd:
            self.dev.update()
            rd.assert_called()

    def test_reboot(self):
        with patch.object(controller.Controller, '_writebus') as wr:
            self.dev.reboot()
            wr.assert_called_once_with(bytes([0x55, 0xFC, 0x1, 0x0, 0x0, 0x0, 0x0, 0xA3, 0x41, 0x95, 0xD2]))
    
    def test_enter_bootloader(self):
        with patch.object(controller.Controller, '_writebus') as wr:
            self.dev.enter_bootloader()
            wr.assert_called_once_with(bytes([0x55, 0xFC, 0x2, 0x0, 0x0, 0x0, 0x0, 0x34, 0xE9, 0x82, 0x9]))

    def test_ping(self):
        self.mock.return_value.read.return_value = bytes([0x55, 0x0, 0x57, 0x73, 0x9D, 0xC6])
        with patch.object(controller.Controller, '_writebus') as wr:
            self.assertTrue(self.dev.ping())
            wr.assert_called_once_with(bytes([0x55, 0x0, 0x57, 0x73, 0x9D, 0xC6]))

    def test_get_status(self):
        self.mock.return_value.read.return_value = bytes([0x55, 0xFC, 0x0, 0x1, 0x0, 0x0, 0x0, 0x1, 0x1, 0x0, 0x15, 0x0, 0x0, 0x0, 0x0, 0xF1, 0x79, 0xD6, 0x6F])
        with patch.object(controller.Controller, '_writebus') as wr:
            st = self.dev.get_board_info()
            wr.assert_called_once_with(bytes([0x55, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2E, 0x26, 0x98, 0x9B]))
            



import unittest
from unittest.mock import patch
from acrome import controller

class TestController(unittest.TestCase):
    def setUp(self):
        patcher = patch("acrome.controller.serial.Serial", autospec=True)
        self.mock = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock.reset_mock()

    def test_crc32(self):
        self.assertEqual(controller.Controller()._crc32([0x55, 0xAA, 123, 321]), bytes([81, 232, 252, 34]))
      
    def test_read(self):
        self.mock.return_value.read.return_value = bytes([0x55, 0x00, 87, 115, 157, 198])
        self.assertEqual(controller.Controller()._readbus(6), bytes([0x55, 0x00, 87, 115, 157, 198]))

    def test_write(self): #Placeholder test since it is just a wrapper
        self.assertTrue(True) 

    def test_reboot(self):
        with patch.object(controller.Controller, '_writebus') as wr:
            controller.Controller().reboot()
            wr.assert_called_once_with(bytes([0x55, 0xFC, 0x1, 0x0, 0x0, 0x0, 0x0, 0xA3, 0x41, 0x95, 0xD2]))
    
    def test_enter_bootloader(self):
        with patch.object(controller.Controller, '_writebus') as wr:
            controller.Controller().enter_bootloader()
            wr.assert_called_once_with(bytes([0x55, 0xFC, 0x2, 0x0, 0x0, 0x0, 0x0, 0x34, 0xE9, 0x82, 0x9]))

    def test_ping(self):
        self.mock.return_value.read.return_value = bytes([0x55, 0x0, 0x57, 0x73, 0x9D, 0xC6])
        with patch.object(controller.Controller, '_writebus') as wr:
            self.assertTrue(controller.Controller().ping())
            wr.assert_called_once_with(bytes([0x55, 0x0, 0x57, 0x73, 0x9D, 0xC6]))

    def test_get_status(self):
        self.mock.return_value.read.return_value = bytes([0x55, 0xFC, 0x0, 0x1, 0x0, 0x0, 0x0, 0x1, 0x1, 0x0, 0x15, 0x0, 0x0, 0x0, 0x0, 0xF1, 0x79, 0xD6, 0x6F])
        with patch.object(controller.Controller, '_writebus') as wr:
            st = controller.Controller().get_board_info()
            wr.assert_called_once_with(bytes([0x55, 0xFC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x2E, 0x26, 0x98, 0x9B]))
        
        self.assertTrue(isinstance(st, dict))
        self.assertEqual(len(st.keys()), len(controller.Controller()._STATUS_KEY_LIST))
        self.assertTrue(list(st.keys()).sort() == controller.Controller()._STATUS_KEY_LIST.sort())
        self.assertTrue(st['EEPROM'] == True)
        self.assertTrue(st['IMU'] == False)
        self.assertTrue(st['Touchscreen Serial'] == True)
        self.assertTrue(st['Touchscreen Analog'] == False)
        self.assertTrue(st['Delta'] == True)
        self.assertTrue(st['Software Version'] == '0.1.0')
        self.assertTrue(st['Hardware Version'] == '1.1.0')

    def test_update(self):
        self.assertRaises(NotImplementedError, controller.Controller().update)